package sun.security.provider.certpath;

import java.security.Provider;
import java.security.cert.X509CRLSelector;
import java.security.cert.CRLException;
import java.security.cert.CRLSelector;
import java.util.Iterator;
import java.util.ArrayList;
import java.io.InputStream;
import java.net.URLConnection;
import java.security.cert.CertStoreException;
import java.security.cert.X509CertSelector;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.security.cert.CertSelector;
import sun.security.x509.GeneralNameInterface;
import sun.security.x509.URIName;
import sun.security.x509.AccessDescription;
import java.security.NoSuchAlgorithmException;
import java.security.InvalidAlgorithmParameterException;
import java.security.cert.CertificateException;
import java.util.Locale;
import java.util.Collections;
import java.security.cert.CertStoreParameters;
import java.security.AccessController;
import sun.security.action.GetIntegerAction;
import sun.security.util.Cache;
import java.security.cert.CertStore;
import java.net.URI;
import java.security.cert.X509CRL;
import java.security.cert.X509Certificate;
import java.util.Collection;
import java.security.cert.CertificateFactory;
import sun.security.util.Debug;
import java.security.cert.CertStoreSpi;

class URICertStore extends CertStoreSpi
{
    private static final Debug debug;
    private static final int CHECK_INTERVAL = 30000;
    private static final int CACHE_SIZE = 185;
    private final CertificateFactory factory;
    private Collection<X509Certificate> certs;
    private X509CRL crl;
    private long lastChecked;
    private long lastModified;
    private URI uri;
    private boolean ldap;
    private CertStoreHelper ldapHelper;
    private CertStore ldapCertStore;
    private String ldapPath;
    private static final int DEFAULT_CRL_CONNECT_TIMEOUT = 15000;
    private static final int CRL_CONNECT_TIMEOUT;
    private static final Cache<URICertStore.URICertStoreParameters, CertStore> certStoreCache;
    
    private static int initializeTimeout() {
        Integer n = (Integer)AccessController.doPrivileged(new GetIntegerAction("com.sun.security.crl.timeout"));
        if (n == null || (int)n < 0) {
            return 15000;
        }
        return (int)n * 1000;
    }
    
    URICertStore(CertStoreParameters params) throws InvalidAlgorithmParameterException, NoSuchAlgorithmException {
        super(params);
        this.certs = Collections.emptySet();
        this.ldap = false;
        if (!(params instanceof URICertStore.URICertStoreParameters)) {
            throw new InvalidAlgorithmParameterException("params must be instanceof URICertStoreParameters");
        }
        this.uri = ((URICertStore.URICertStoreParameters)params).uri;
        if (this.uri.getScheme().toLowerCase(Locale.ENGLISH).equals("ldap")) {
            this.ldap = true;
            this.ldapHelper = CertStoreHelper.getInstance("LDAP");
            this.ldapCertStore = this.ldapHelper.getCertStore(this.uri);
            this.ldapPath = this.uri.getPath();
            if (this.ldapPath.charAt(0) == '/') {
                this.ldapPath = this.ldapPath.substring(1);
            }
        }
        try {
            this.factory = CertificateFactory.getInstance("X.509");
        }
        catch (CertificateException ex) {
            throw new RuntimeException();
        }
        return;
    }
    
    static synchronized CertStore getInstance(URICertStore.URICertStoreParameters uriCertStoreParameters) throws NoSuchAlgorithmException, InvalidAlgorithmParameterException {
        if (URICertStore.debug != null) {
            URICertStore.debug.println("CertStore URI:" + uriCertStoreParameters.uri);
        }
        CertStore certStore = (CertStore)URICertStore.certStoreCache.get(uriCertStoreParameters);
        if (certStore == null) {
            certStore = new URICertStore.UCS(new URICertStore(uriCertStoreParameters), null, "URI", uriCertStoreParameters);
            URICertStore.certStoreCache.put(uriCertStoreParameters, certStore);
        }
        else if (URICertStore.debug != null) {
            URICertStore.debug.println("URICertStore.getInstance: cache hit");
        }
        return certStore;
    }
    
    static CertStore getInstance(AccessDescription accessDescription) {
        if (!accessDescription.getAccessMethod().equals(AccessDescription.Ad_CAISSUERS_Id)) {
            return null;
        }
        GeneralNameInterface name = accessDescription.getAccessLocation().getName();
        if (!(name instanceof URIName)) {
            return null;
        }
        URI uri = ((URIName)name).getURI();
        try {
            return getInstance(new URICertStoreParameters(uri));
        }
        catch (Exception obj) {
            if (URICertStore.debug != null) {
                URICertStore.debug.println("exception creating CertStore: " + obj);
                obj.printStackTrace();
            }
            return null;
        }
    }
    
    @Override
    public synchronized Collection<X509Certificate> engineGetCertificates(CertSelector certSelector) throws CertStoreException {
        if (this.ldap) {
            X509CertSelector x509CertSelector = (X509CertSelector)certSelector;
            X509CertSelector wrap;
            try {
                wrap = this.ldapHelper.wrap(x509CertSelector, x509CertSelector.getSubject(), this.ldapPath);
            }
            catch (IOException cause) {
                throw new CertStoreException(cause);
            }
            return this.ldapCertStore.getCertificates(wrap);
        }
        long currentTimeMillis = System.currentTimeMillis();
        if (currentTimeMillis - this.lastChecked < 30000L) {
            if (URICertStore.debug != null) {
                URICertStore.debug.println("Returning certificates from cache");
            }
            return getMatchingCerts(this.certs, certSelector);
        }
        else {
            this.lastChecked = currentTimeMillis;
            try {
                URLConnection openConnection = this.uri.toURL().openConnection();
                if (this.lastModified != 0L) {
                    openConnection.setIfModifiedSince(this.lastModified);
                }
                long lastModified = this.lastModified;
                try (final InputStream inputStream = openConnection.getInputStream()) {
                    this.lastModified = openConnection.getLastModified();
                    if (lastModified != 0L) {
                        if (lastModified == this.lastModified) {
                            if (URICertStore.debug != null) {
                                URICertStore.debug.println("Not modified, using cached copy");
                            }
                            return getMatchingCerts(this.certs, certSelector);
                        }
                        else if (openConnection instanceof HttpURLConnection && ((HttpURLConnection)openConnection).getResponseCode() == 304) {
                            if (URICertStore.debug != null) {
                                URICertStore.debug.println("Not modified, using cached copy");
                            }
                            return getMatchingCerts(this.certs, certSelector);
                        }
                    }
                    if (URICertStore.debug != null) {
                        URICertStore.debug.println("Downloading new certificates...");
                    }
                    this.certs = this.factory.generateCertificates(inputStream);
                }
                return getMatchingCerts(this.certs, certSelector);
            }
            catch (IOException | CertificateException ex) {
                if (URICertStore.debug != null) {
                    URICertStore.debug.println("Exception fetching certificates:");
                    ex.printStackTrace();
                }
                this.lastModified = 0L;
                return this.certs = Collections.emptySet();
            }
        }
    }
    
    private static Collection<X509Certificate> getMatchingCerts(Collection<X509Certificate> collection, CertSelector certSelector) {
        if (certSelector == null) {
            return collection;
        }
        ArrayList list = new ArrayList(collection.size());
        for (X509Certificate x509Certificate : collection) {
            if (certSelector.match(x509Certificate)) {
                list.add(x509Certificate);
            }
            else {
                continue;
            }
        }
        return list;
    }
    
    @Override
    public synchronized Collection<X509CRL> engineGetCRLs(CRLSelector crlSelector) throws CertStoreException {
        if (this.ldap) {
            X509CRLSelector x509CRLSelector = (X509CRLSelector)crlSelector;
            X509CRLSelector wrap;
            try {
                wrap = this.ldapHelper.wrap(x509CRLSelector, null, this.ldapPath);
            }
            catch (IOException cause2) {
                throw new CertStoreException(cause2);
            }
            try {
                return this.ldapCertStore.getCRLs(wrap);
            }
            catch (CertStoreException ex) {
                throw new PKIX.CertStoreTypeException("LDAP", ex);
            }
        }
        else {
            long currentTimeMillis = System.currentTimeMillis();
            if (currentTimeMillis - this.lastChecked < 30000L) {
                if (URICertStore.debug != null) {
                    URICertStore.debug.println("Returning CRL from cache");
                }
                return getMatchingCRLs(this.crl, crlSelector);
            }
            else {
                this.lastChecked = currentTimeMillis;
                try {
                    URLConnection openConnection = this.uri.toURL().openConnection();
                    if (this.lastModified != 0L) {
                        openConnection.setIfModifiedSince(this.lastModified);
                    }
                    long lastModified = this.lastModified;
                    openConnection.setConnectTimeout(URICertStore.CRL_CONNECT_TIMEOUT);
                    try (final InputStream inputStream = openConnection.getInputStream()) {
                        this.lastModified = openConnection.getLastModified();
                        if (lastModified != 0L) {
                            if (lastModified == this.lastModified) {
                                if (URICertStore.debug != null) {
                                    URICertStore.debug.println("Not modified, using cached copy");
                                }
                                return getMatchingCRLs(this.crl, crlSelector);
                            }
                            else if (openConnection instanceof HttpURLConnection && ((HttpURLConnection)openConnection).getResponseCode() == 304) {
                                if (URICertStore.debug != null) {
                                    URICertStore.debug.println("Not modified, using cached copy");
                                }
                                return getMatchingCRLs(this.crl, crlSelector);
                            }
                        }
                        if (URICertStore.debug != null) {
                            URICertStore.debug.println("Downloading new CRL...");
                        }
                        this.crl = (X509CRL)this.factory.generateCRL(inputStream);
                    }
                    return getMatchingCRLs(this.crl, crlSelector);
                }
                catch (IOException | CRLException cause) {
                    if (URICertStore.debug != null) {
                        URICertStore.debug.println("Exception fetching CRL:");
                        cause.printStackTrace();
                    }
                    this.lastModified = 0L;
                    this.crl = null;
                    throw new PKIX.CertStoreTypeException("URI", new CertStoreException(cause));
                }
            }
        }
    }
    
    private static Collection<X509CRL> getMatchingCRLs(X509CRL o, CRLSelector crlSelector) {
        if (crlSelector == null || (o != null && crlSelector.match(o))) {
            return Collections.singletonList(o);
        }
        return Collections.emptyList();
    }
    
    static {
        debug = Debug.getInstance("certpath");
        CRL_CONNECT_TIMEOUT = initializeTimeout();
        certStoreCache = Cache.newSoftMemoryCache(185);
    }
    
    static class URICertStoreParameters implements CertStoreParameters
    {
        private final URI uri;
        private volatile int hashCode;
        
        URICertStoreParameters(URI uri) {
            this.hashCode = 0;
            this.uri = uri;
        }
        
        @Override
        public boolean equals(Object o) {
            return o instanceof URICertStore.URICertStoreParameters && this.uri.equals(((URICertStore.URICertStoreParameters)o).uri);
        }
        
        @Override
        public int hashCode() {
            if (this.hashCode == 0) {
                this.hashCode = 37 * 17 + this.uri.hashCode();
            }
            return this.hashCode;
        }
        
        @Override
        public Object clone() {
            try {
                return super.clone();
            }
            catch (CloneNotSupportedException cause) {
                throw new InternalError(cause.toString(), cause);
            }
        }
    }
    
    private static class UCS extends CertStore
    {
        protected UCS(CertStoreSpi storeSpi, Provider provider, String type, CertStoreParameters params) {
            super(storeSpi, provider, type, params);
        }
    }
}
